package applications.links;

import applications.collages.collage;
import applications.collages.collageLike;

import java.util.Enumeration;
import java.util.LinkedList;
import java.awt.geom.*;

public class link extends collage implements collageLike, Cloneable {
  
  private LinkedList<AffineTransform> trans = new LinkedList<AffineTransform>();
  private LinkedList<collage> col = new LinkedList<collage>();
  private collage centre;

  public link(collage c) {
    col.add(centre = c);
  }
  
  public static link emptyLink() {
    link result = new link(new collage());
    result.trans.add(new AffineTransform());
    return result;
  }
  
  public static link push() {
    link result = emptyLink();
    result.col.addLast(new collage());
    result.trans.addLast(new AffineTransform());
    return result;
  }
  
  public static link pop() {
    link result = emptyLink();
    result.col.addFirst(new collage());
    result.trans.addFirst(new AffineTransform());
    return result;
  }
  
  public void setTransform(AffineTransform t) {
    if (trans.size() != 0) throw new InternalError("setTransform applied to already initialized link");
    trans.add(t);
  }
  
  public void concatWith(link l) {
    while (col.getLast() != centre && l.col.getFirst() != l.centre) {
      collage c1 = col.getLast();
      collage c2 = l.col.getFirst();
      AffineTransform t = trans.getLast();
      col.removeLast();
      trans.removeLast();
      l.col.removeFirst();
      l.trans.removeFirst();
      c2.transform(t);
      col.getLast().unionWith(c1);
      col.getLast().unionWith(c2);
    }
    if (l.col.getFirst()!=l.centre) {
      l.col.getFirst().transform(trans.getLast());
      centre.unionWith(l.col.getFirst());
      l.col.removeFirst();
      l.trans.removeFirst();
      col.addAll(l.col);
      trans.addAll(l.trans);
      centre = l.centre;
    }
    else {
      AffineTransform t = (AffineTransform) trans.getLast().clone();
      col.getLast().unionWith(l.centre);
      trans.set(trans.size()-1,l.trans.getFirst());
      while (true) {
        l.col.getFirst().transform(t);
        l.col.removeFirst();
        l.trans.getFirst().preConcatenate(t);
        l.trans.removeFirst();
        if (l.col.isEmpty()) break;
        col.addLast(l.col.getFirst());
        trans.addLast(l.trans.getFirst());
      }
    }
  }
  
  public Enumeration parts() {
    return new partEnumeration();
  }
  
  public int numParts() {
    int result = 0;
    for (int i = 0; i < col.size(); i++) result += col.get(i).numParts();
    return result;
  }
  
  public Rectangle2D.Double bounds() {
    Rectangle2D.Double result = col.getFirst().bounds();
    for (int i = 1; i < col.size(); i++) result.add(col.get(i).bounds());
    return result;
  }
  
  public String toString() {
    return "(" + col.toString() + "<" + centre + ">, " + trans.toString() + ")";
  }
  
  public Object clone() {
    link result = (link)super.clone();
    result.col = (LinkedList<collage>) col.clone();
    result.trans = (LinkedList<AffineTransform>)trans.clone();
    result.cloneLists();
    return result;
  }
  
  private void cloneLists() {
    for (int i = 0; i < col.size(); i++) {
      trans.set(i, (AffineTransform) trans.get(i).clone());
      if (centre == col.set(i, (collage) col.get(i).clone())) {
        centre = col.get(i);
      }
    }
  }
  
  private class partEnumeration implements Enumeration {
    private int curr = 0;
    private Enumeration currEnum = col.getFirst().parts();

    public boolean hasMoreElements() {
      while (true) {
        if (currEnum.hasMoreElements()) return true;
        if (++curr < col.size()) currEnum = col.get(curr).parts();
        else return false;
      }
    }
    
    public Object nextElement() {
      if (hasMoreElements()) return currEnum.nextElement();
      throw new InternalError("link.parts().nextElement() called when there was none");
    }
  }

}
