/* Generated By:JavaCC: Do not edit this line. dbuTransducerParser.java */
package parsers;

import java.util.*;
import terms.*;

/** The parser for deterministic bottom-up tree transducers.
  * Expects to find a six-tuple consisting of the input signature, output signatue,
  * state set (disjoint with both the input and output signature), rule set, and set of final states.
  @see finiteSignatureParser
  */
public class dbuTransducerParser implements dbuTransducerParserConstants {

  private ASCII_CharStream inputStream = null;
  protected rulesParser rules;

  public dbuTransducerParser(ASCII_CharStream input) {
    this(new dbuTransducerParserTokenManager(input));
    rules = new rulesParser(input);
    inputStream = input;
  }

  public Vector lhsV = new Vector();
  public Vector rhsV = new Vector();
  public fixedRankSignature states;
  public fixedRankSignature finalStates;
  public finiteSignature in;
  public finiteSignature out;

  private class rulesParser extends trsParser {

    public rulesParser(ASCII_CharStream input) {
      super(input);
    }

    protected void insertRule(term lhs, term rhs, double wght) throws ParseException {
      verifyRule(lhs, rhs);
      if (lhsV.contains(lhs)) throw new ParseException("Left-hand side occurs twice.");
      lhsV.add(lhs);
      rhsV.add(rhs);
    }
  }

  protected void verifyRule(term lhs, term rhs) throws ParseException {
    verifyLhs(lhs);
    verifyRhs(rhs, lhs.topSymbol().rank()+1);
  }

  private void verifyLhs(term lhs) throws ParseException {
    symbol sym = lhs.topSymbol();
    if (!in.contains(sym)) throw new ParseException("Topmost symbol " + sym + " of left-hand side not an input symbol.");
    for (int i = 0; i < sym.rank(); i++) {
      term s = lhs.subterm(i);
      if (!states.contains(s.topSymbol()) ||
          !(s.subterm(0).topSymbol() instanceof variable) ||
          ((variable)s.subterm(0).topSymbol()).index() != i+1) {
        throw new ParseException("Malformed left-hand side.");
      }
    }
  }

  private void verifyRhs(term rhs, int maxIndex) throws ParseException {
    if (states.contains(rhs.topSymbol())) verifyRhsSubtree(rhs.subterm(0), maxIndex);
    else throw new ParseException("Root symbol of rhs not a state");
  }

  private void verifyRhsSubtree(term rhs, int maxIndex) throws ParseException {
    symbol top = rhs.topSymbol();
    if (out.contains(top)) {
      for (int i = 0; i < top.rank(); i++) {
        verifyRhsSubtree(rhs.subterm(i),maxIndex);
      }
    }
    else if (top instanceof variable) {
      int i = ((variable)top).index();
      if (i == 0 || i > maxIndex) throw new ParseException("Variable in rhs does not appear in lhs.");
    }
    else throw new ParseException("Symbol in rhs is neither an output symbol nor a variable.");
  }

/** Parse a deterministic bottom-up tree transducer.
  * Expects to find a six-tuple consisting of the input signature, output signatue,
  * state set (disjoint with both the input and output signature, rule set, and set of final states.
  * @exception ParseException if an error occurs
  */
  final public void dbuTransducer() throws ParseException {
    jj_consume_token(LEFT_PAR);
      in = new finiteSignature();
      in.parse(inputStream);
    jj_consume_token(COMMA);
      out = new finiteSignature();
      out.parse(inputStream);
    jj_consume_token(COMMA);
      states = new fixedRankSignature(1);
      states.parse(inputStream);
      if (!states.disjointWith(in) || !states.disjointWith(out)) {
        {if (true) throw new ParseException("set of states must be disjoint with input and output signatures");}
      }
    jj_consume_token(COMMA);
      rules.rules();
    jj_consume_token(COMMA);
      finalStates = new fixedRankSignature(1);
      finalStates.parse(inputStream);
      if (!states.contains(finalStates)) {
        {if (true) throw new ParseException("Set of final states not a subset of the set of states");}
      }
  }

  public dbuTransducerParserTokenManager token_source;
  ASCII_CharStream jj_input_stream;
  public Token token, jj_nt;
  private int jj_ntk;
  private int jj_gen;
  final private int[] jj_la1 = new int[0];
  final private int[] jj_la1_0 = {};

  public dbuTransducerParser(java.io.InputStream stream) {
    jj_input_stream = new ASCII_CharStream(stream, 1, 1);
    token_source = new dbuTransducerParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  public void ReInit(java.io.InputStream stream) {
    jj_input_stream.ReInit(stream, 1, 1);
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  public dbuTransducerParser(java.io.Reader stream) {
    jj_input_stream = new ASCII_CharStream(stream, 1, 1);
    token_source = new dbuTransducerParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  public void ReInit(java.io.Reader stream) {
    jj_input_stream.ReInit(stream, 1, 1);
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  public dbuTransducerParser(dbuTransducerParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  public void ReInit(dbuTransducerParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 0; i++) jj_la1[i] = -1;
  }

  final private Token jj_consume_token(int kind) throws ParseException {
    Token oldToken;
    if ((oldToken = token).next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    if (token.kind == kind) {
      jj_gen++;
      return token;
    }
    token = oldToken;
    jj_kind = kind;
    throw generateParseException();
  }

  final public Token getNextToken() {
    if (token.next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    jj_gen++;
    return token;
  }

  final public Token getToken(int index) {
    Token t = token;
    for (int i = 0; i < index; i++) {
      if (t.next != null) t = t.next;
      else t = t.next = token_source.getNextToken();
    }
    return t;
  }

  final private int jj_ntk() {
    if ((jj_nt=token.next) == null)
      return (jj_ntk = (token.next=token_source.getNextToken()).kind);
    else
      return (jj_ntk = jj_nt.kind);
  }

  private java.util.Vector jj_expentries = new java.util.Vector();
  private int[] jj_expentry;
  private int jj_kind = -1;

  final public ParseException generateParseException() {
    jj_expentries.removeAllElements();
    boolean[] la1tokens = new boolean[9];
    for (int i = 0; i < 9; i++) {
      la1tokens[i] = false;
    }
    if (jj_kind >= 0) {
      la1tokens[jj_kind] = true;
      jj_kind = -1;
    }
    for (int i = 0; i < 0; i++) {
      if (jj_la1[i] == jj_gen) {
        for (int j = 0; j < 32; j++) {
          if ((jj_la1_0[i] & (1<<j)) != 0) {
            la1tokens[j] = true;
          }
        }
      }
    }
    for (int i = 0; i < 9; i++) {
      if (la1tokens[i]) {
        jj_expentry = new int[1];
        jj_expentry[0] = i;
        jj_expentries.addElement(jj_expentry);
      }
    }
    int[][] exptokseq = new int[jj_expentries.size()][];
    for (int i = 0; i < jj_expentries.size(); i++) {
      exptokseq[i] = (int[])jj_expentries.elementAt(i);
    }
    return new ParseException(token, exptokseq, tokenImage);
  }

  final public void enable_tracing() {
  }

  final public void disable_tracing() {
  }

}
